#include <stdio.h> 
#include <avr/io.h>
#include <avr/interrupt.h>
#include "bool.h"
#include "SevenSeg.h"
#include "trtSettings.h"
#include "trtLib.h"

// debouncing variables and constants
#define NoPush 0
#define Pushed 1
#define Maybe 2
volatile char incrHourState, incrMinState;

#define SNOOZE_TIME 9 //in minutes

// count DDS ISRs to generate second time base
unsigned int count = 0, debounceCount = 0, motionCount = 0;
#define DEBOUNCE_EVERY 1875
#define MOTION_LENGTH 62500

volatile bool beepCycle = false;
volatile bool alarmOn = false; //true if the alarm switch is on
volatile bool toneOn = false; //true if the alarm tone should be playing
volatile bool showAlarmTime = false; //true if alarm set is asserted

// Current time
unsigned char hour = 0, minute = 3, second = 0;

// Alarm time
unsigned char aHour = 0, aMinute = 8;
unsigned char sHour, sMinute; // "effective" alarm time w/ snooze

void displayTime(char hour, char minute) {
	char d1 = hour/10;
	if (d1 == 0) d1 = -1;
	seven_seg_setDigits(d1,hour%10,minute/10,minute%10);
}

// ISR - runs at 62.5 kHz - generates time base
ISR (TIMER0_OVF_vect) {
	if (beepCycle)
		PORTB = PORTB | 0b00000001;
	else PORTB = PORTB & 0b11111110;

    // increment time variables
    count++; debounceCount++;
	if (motionCount > 0) motionCount--;
	if (count == 31250) {
		beepCycle = toneOn && !beepCycle;
		seven_seg_toggleColon();
	}
	//increment clock time
    if (count >= 62500) {
        count = 0;
		beepCycle = toneOn && !beepCycle;
        second++;
		if (second > 59) {
			second = 0;
			minute++;
			if (minute > 59) {
				minute = 0;
				hour = (hour + 1) % 24;
			}
		}
		if (sHour == hour && sMinute == minute && second == 0)
			toneOn = alarmOn; //turn on alarm tone if alarm switch on
		if (!showAlarmTime) // display the new time
			displayTime(hour,minute);
    }
}

// called when snooze button pressed
void snooze() {
	toneOn = false;
	sMinute = minute + SNOOZE_TIME;
	if (sMinute > 59) {
		sMinute = sMinute % 60;
		sHour = (sHour + 1) % 24;
	}
}

// called when alarm on switch goes from OFF to ON
void turnAlarmOn() {
	sHour = aHour;
	sMinute = aMinute;
	alarmOn = true;
	seven_seg_setUpperDot(true);
}

// called when alarm on switch goes from OFF to ON
void turnAlarmOff() {
	alarmOn = toneOn = false;
	seven_seg_setUpperDot(false);
}

// called when the increment hour button is pressed
void incrMinute() {
	if (!(PINA & 0x01)) { //set time asserted
		minute = (minute + 1) % 60;
		displayTime(hour,minute);
	} else if (!(PINA & 0x02)) { //set alarm asserted
		aMinute = (aMinute + 1) % 60;
		if (alarmOn) 
			sMinute = aMinute;
		displayTime(aHour,aMinute);
	}
}

// called when the increment hour button is pressed
void incrHour() {
	if (!(PINA & 0x01)) { //set time asserted
		hour = (hour + 1) % 24;
		displayTime(hour,minute);
	} else if (!(PINA & 0x02)) { //set alarm asserted
		aHour = (aHour + 1) % 24;
		if (alarmOn) 
			sHour = aHour;
		displayTime(aHour,aMinute);
	}
}

// Debounces the incr. hour and incr. minute buttons
void debounce() {
	switch (incrHourState) {
	case NoPush: 
		if (!(PINA & 0x08)) //pushed
			incrHourState = Maybe;
		break;
	case Maybe:
		if (!(PINA & 0x08)) {
			incrHourState = Pushed;
			incrHour();
		}
		else incrHourState = NoPush;
		break;
	case Pushed:  
		if (PINA & 0x08)
			incrHourState = Maybe;
		break;
	}

	switch (incrMinState) {
	case NoPush: 
		if (!(PINA & 0x04))
			incrMinState = Maybe;
		break;
	case Maybe:
		if (!(PINA & 0x04)) {
			incrMinState = Pushed;
			incrMinute();
		}
		else incrMinState = NoPush;
		break;
	case Pushed:  
		if (PINA & 0x04)
			incrMinState = Maybe;
		break;
	}
}

// Initialize the timers, ports, and interrupts 
void initialize() {
    TIMSK0 = (1<<TOIE0); // turn on timer 0 overflow ISR
    TCCR0B = 1; // timer 0 runs at full rate
	sHour = aHour;
	sMinute = aMinute;

	DDRA = 0; //use port A as inputs
	DDRB = 0x03; //make b output
    PORTB = 0xff;
	sei(); // turn on master interrupts
} 

int main () {
	initialize();
	seven_seg_init();
	seven_seg_setUpperDot(alarmOn);
	trtInitKernel(100); //100 bytes for the idle task stack
  	trtCreateTask(seven_seg_run, 200, SECONDS2TICKS(0.001),
  		SECONDS2TICKS(0.002), 0);
	
	while (1) {
		// Alarm set button
		if (!(PINA & 0x02)) {
			showAlarmTime = true;
			displayTime(aHour,aMinute);
		} else {
			showAlarmTime = false;
			displayTime(hour,minute);
		}
		// Alarm On/Off switch
		if (PINA & 0x20 && alarmOn)
			turnAlarmOff();
		else if (!(PINA & 0x20) && !alarmOn)
			turnAlarmOn();
		// Snooze button press
		if (!(PINA & 0x10) && toneOn)
			snooze();
		// Snooze proximity sensor
		if ((!(PINA & 0x40) || !(PINA & 0x80))&& toneOn && motionCount == 0)
			motionCount = MOTION_LENGTH;
		else if (!toneOn) motionCount = 0;
		// Drive motor control
		if (motionCount > 0) PORTB |= 0x02;
		else PORTB &= 0xfd;
		// Possibly call debouncer
		if (debounceCount > DEBOUNCE_EVERY) {
			debounceCount = 0;
			debounce();
		}
	}
}
